/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2002 Ian Zimmerman <itz@speakeasy.org> */

static const char rcsid[] = "$Id: handlers.c,v 1.2 2002/06/17 23:40:10 itz Exp $ ";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <sys/time.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <assert.h>

#include "gim_event.h"
#include "exc.h"
#include "gim_common.h"

int
gim_event_select (int fd, gim_handler h, struct gim_handler_data_t* hdata,
                  int (*getcfun)(void* p), void* arg)
{
  int c;
  int flags, newflags;
  fd_set readset;
  int maxfd;
  int status;
  struct gim_event_t ev;

  exc_try {

    if (hdata->morekeys > 0) {
      hdata->is_fake = 1;
      c = (*h)(0, hdata);
      exc_ereturn2(int, c);
    } /*if*/

    /* loop until we get something */
    while (1) {
      FD_ZERO(&readset);
      maxfd = 0;
      
      /* aaah, stdio buffering */
      if (0 > (flags = fcntl(STDIN_FILENO, F_GETFL))) exc_eraise2(unix_error, errno);
      newflags = flags | O_NONBLOCK;
      if (0 > fcntl(STDIN_FILENO, F_SETFL, newflags)) exc_eraise2(unix_error, errno);

      errno = 0;
      exc_try {
        if (EOF == (c = (*getcfun)(arg))) {
          switch (errno) {
          case EWOULDBLOCK:
            FD_SET(STDIN_FILENO, &readset);
            if (STDIN_FILENO + 1 > maxfd) maxfd = STDIN_FILENO + 1;
            break;
          case 0:
            /* a real EOF */
            break;
          default:
            exc_eraise2(unix_error, errno);
          } /*switch*/
        } /*if*/
      }
      exc_finally {
        fcntl(STDIN_FILENO, F_SETFL, flags);
      } exc_end;

      if (!FD_ISSET(STDIN_FILENO, &readset)) {
        hdata->is_fake = 0;
        errno = 0;
        exc_ereturn2(int, c);
      } /*if*/

      FD_SET(fd, &readset);
      if (fd + 1 > maxfd) maxfd = fd + 1;
      /* back to "loop until we get something" */
      if (0 <= (status = select(maxfd, &readset, 0, 0, 0)) || errno != EINTR) break;
    } /*while*/

    if (status < 0) exc_eraise2(unix_error, errno);
    assert(status > 0);
      
    if (FD_ISSET(STDIN_FILENO, &readset)) {
      hdata->is_fake = 0;
      c = (*getcfun)(arg);
      exc_ereturn2(int, c);
    } else if (FD_ISSET(fd, &readset)) {
      hdata->is_fake = 1;
      gim_get_event_internal(fd, &ev);
      c = (*h)(&ev, hdata);
      exc_ereturn2(int, c);
    } else {
      assert(0);
    } /*if*/
  }
  exc_except(unix_error) {
    errno = *(int *)exc_info;
    exc_ereturn2(int, EOF);
  }
  exc_except(exc_any) {
    fprintf(stderr, "%s: %d: %s\n", exc_file, exc_line, exc_name);
    exc_ereturn2(int, EOF);
  } exc_end;
}

static int
my_getchar (void* p)
{
  return getchar();
}

int 
gim_getchar (int fd, gim_handler h, struct gim_handler_data_t* hdata)
{
  return gim_event_select(fd, h, hdata, my_getchar, 0);
}

void
gim_init_region (struct gim_region_t* p_reg)
{
  /* make really sure it's ignored :) */
  p_reg->x_min = 0; p_reg->x_max = -1; p_reg->y_min = 0; p_reg->y_max = -1;
  p_reg->min_mod = ~0; p_reg->max_mod = 0;
  p_reg->handle = 0;
  p_reg->handler_fun = 0;
  p_reg->extra_data = 0;
  p_reg->active = 0;
  p_reg->prev = p_reg->next = p_reg;
}

struct gim_region_t*
gim_alloc_region (void) 
{
  struct gim_region_t* pr;

  pr = (struct gim_region_t*) malloc(sizeof(struct gim_region_t));
  if (pr == 0) return 0;
  gim_init_region(pr);
  return pr;
}

void
gim_put_region (struct gim_region_t* reg, struct gim_region_t* after)
{
  struct gim_region_t* prev;
  struct gim_region_t* next;

  /* fisrt remove it from current position, if needed */
  prev = reg->prev; next = reg->next;
  prev->next = next;
  next->prev = prev;

  /* then insert it in the new position */
  if (after != 0 && after != reg) {
    reg->prev = after;
    reg->next = after->next;
    after->next->prev = reg;
    after->next = reg;
  } else {
    reg->next = reg->prev = reg;
  } /*if*/
}

#define IS_SUBSET(a,b) (((a)|(b)) == (b))

int
gim_region_handler (struct gim_event_t* p_ev, struct gim_handler_data_t* hdata)
{
  struct gim_region_t* top;
  struct gim_region_t* p_reg;
  struct gim_region_t* p_active;
  struct gim_region_t* p_new;
  struct gim_event_t synth_ev;

  top = (struct gim_region_t*)(hdata->extra_data);

  /* find active region, if any.  if the event pointer is null, the active
     region's handler will be called, to allow for morekeys processing */

  p_active = 0;
  for (p_reg = top->next; p_reg != top; p_reg = p_reg->next) {
    if (p_reg->active) {
      p_active = p_reg;
      break;
    } /*if*/
  } /*for*/

  if (p_ev == 0 && p_active != 0 && p_active->handler_fun != 0) {
    hdata->extra_data = p_active->extra_data;
    return (*p_active->handler_fun)(0, hdata);
  } /*if*/

  assert(p_ev != 0);
  p_new = p_active;
  if (0 != (p_ev->kind & (gim_move_ev | gim_down_ev))) {
    for (p_reg = top->next; p_reg != top; p_reg = p_reg->next) {
      if (p_reg->x_min <= p_ev->x && p_ev->x < p_reg->x_max &&
          p_reg->y_min <= p_ev->y && p_ev->y < p_reg->y_max &&
          IS_SUBSET(p_reg->min_mod, p_ev->modifiers) &&
          IS_SUBSET(p_ev->modifiers, p_reg->max_mod)) {
        p_new = p_reg;
        break;
      } /*if*/
    } /*for*/
  } /*if*/

  /* pass a leave event to the active region, and deactivate it */
  if (p_new != p_active) {
    synth_ev.x = p_ev->x;
    synth_ev.y = p_ev->y;
    synth_ev.z = p_ev->z;
    synth_ev.state = p_ev->state;
    synth_ev.modifiers = p_ev->modifiers;
    synth_ev.time = p_ev->time;
    synth_ev.vt = p_ev->vt;
    if (0 != (p_ev->kind & gim_position_ev)) {
      synth_ev.u.d.x = p_ev->u.d.x;
      synth_ev.u.d.y = p_ev->u.d.y;
      synth_ev.u.d.z = p_ev->u.d.z;
      synth_ev.u.d.margin = p_ev->u.d.margin;
    } else if (0 != (p_ev->kind & gim_button_ev)) {
      synth_ev.u.b.which = p_ev->u.b.which;
      synth_ev.u.b.repeat = p_ev->u.b.repeat;
    } else {
      assert(0);
    } /*if*/
    if (p_active != 0 && 0 != (p_active->handle & gim_leave_ev)) {
      synth_ev.kind = gim_leave_ev;
      p_active->active = 0;

    /* FIXME: return value of handlers is ignored for enter and leave */
      hdata->extra_data = p_active->extra_data;
      (void)((*p_active->handler_fun)(&synth_ev, hdata));
    }  /*if*/
    if (p_new != 0 && 0 != (p_new->handle & gim_enter_ev)) {
      synth_ev.kind = gim_enter_ev;
      p_new->active = 1;

    /* FIXME: return value of handlers is ignored for enter and leave */
      hdata->extra_data = p_new->extra_data;
      (void)((*p_new->handler_fun)(&synth_ev, hdata));
    } /*if*/
  } /*if*/

  /* finally, pass the real event to the correct handler */
  if (p_new != 0 && 0 != (p_new->handle & p_ev->kind)) {
    hdata->extra_data = p_new->extra_data;
    return (*p_new->handler_fun)(p_ev, hdata);
  } else {
    return 0;
  } /*if*/
}
