/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static char rcsid[] = "$Id: gim.c,v 1.33 2002/09/11 20:23:48 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ggi/gii.h>
#include <ggi/gii-unix.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <unistd.h>
#include <paths.h>
#include <syslog.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <getopt.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <dlfcn.h>

#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#endif 

#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif

#include "dispatch.h"
#include "sockets.h"
#include "connections.h"
#include "linuxonly.h"
#include "options.h"
#include "cook.h"
#include "exc.h"
#include "gim_common.h"
#include "bitsets.h"

gii_cmddata_getdevinfo gii_di;

typedef int (*repeat_event_t)(const gii_event* pev);
typedef int (*repeat_init_t)(void);

static volatile sig_atomic_t got_sig;

#define WAIT_TEXT_TIMEOUT 2     /* seconds to wait before rechecking text/graphics mode */

declare_bits(word_chars, UCHAR_MAX);
declare_bits(ignored, MAX_NR_CONSOLES+1);

static void
handle_sig (int signum)
{
  got_sig = signum;
}

static void
event_loop (gii_input_t giin, int unix_fd, repeat_event_t rpt)
{
  static fd_set client_set;
  static int max_client = 0;
  fd_set allfd_set;
  int max_fd;
  gii_event ev;
  sigset_t allsig_set;
  int i;
  gii_event_mask m;
  struct timeval tv;
  int need_reopen;

  if (max_client == 0) FD_ZERO(&client_set);
  
  sigemptyset(&allsig_set);
  sigaddset(&allsig_set, SIGINT);
  sigaddset(&allsig_set, SIGTERM);
  sigaddset(&allsig_set, SIGHUP);
  sigaddset(&allsig_set, SIGPIPE);

  need_reopen = 0;
  while(1) {
    /* check for signals first */
    sigprocmask(SIG_BLOCK, &allsig_set, 0);
    if (got_sig) {
      if (got_sig == SIGPIPE) {
        /* nobody's reading repeater output, just continue */
        got_sig = 0;
        sigprocmask(SIG_UNBLOCK, &allsig_set, 0);
        continue;               /* back to while */
      } /*if*/
      syslog(LOG_INFO,"terminating on signal %d: %s",
             got_sig, strsignal(got_sig));
      exc_eraise2(async_signal, got_sig);
    }  /*if*/

    sigprocmask(SIG_UNBLOCK, &allsig_set, 0);

    /* setup select() arguments */
    allfd_set = client_set;
    max_fd = max_client;
    if (unix_fd != -1) {
      FD_SET(unix_fd, &allfd_set);
    } /*if*/
    if (unix_fd + 1 > max_fd) {
      max_fd = unix_fd + 1;
    } /*if*/

    if (is_text_mode() && !test_bit(ignored, current_vt())) {
      if (need_reopen) {
        syslog(LOG_INFO, "detected return to text mode, reopening gii stream");
        return;
      } /*if*/
      m = emPointer;
      i = giiEventSelect(giin, &m, max_fd, &allfd_set, 0, 0, 0);
    } else {
      if (!need_reopen) {
        syslog(LOG_INFO, "detected graphics mode, ignoring mouse events");
        need_reopen = 1;
      } /*if*/
      tv.tv_sec = WAIT_TEXT_TIMEOUT;
      tv.tv_usec = 0;
      if (rpt == 0) {
        m = emZero;
        i = select(max_fd, &allfd_set, 0, 0, &tv);
      } else {
        m = emPointer;
        i = giiEventSelect(giin, &m, max_fd, &allfd_set, 0, 0, &tv);
      } /*if*/
    } /*if*/
    if (i == -1 && errno != EINTR) {
      int my_errno = errno;
      syslog(LOG_ERR, "select() failed: %m");
      exc_eraise2(unix_error, my_errno);
    } else if (i == -1 && errno == EINTR) {
      continue;                 /* go to signal check above */
    } /*if*/

    if (m != emZero) {
      if (giiEventRead(giin, &ev, emPointer) == 0) {
        syslog(LOG_ERR, "giiEventRead() failed, is the mouse connected?");
        exc_eraise(gii_error);
      } /*if*/
      if (is_text_mode() && !test_bit(ignored, current_vt())) {
        dispatch_event(&ev);
      } else if (rpt != 0) {
        if (0 > (*rpt)(&ev) && errno != EINTR) {
          int my_errno = errno;
          syslog(LOG_ERR, "repeater failed: %m");
          exc_eraise2(unix_error, my_errno);
        } /*if*/
      } /*if*/
    } /*if*/
    for (i = 0; i < max_client; i = i + 1) {
      if (FD_ISSET(i, &client_set) && FD_ISSET(i, &allfd_set)) {
        /* got request from client */
        handle_client_request(i, &max_client, &client_set);
      } /*if*/
    } /*for*/
    if (unix_fd != -1 && FD_ISSET(unix_fd, &allfd_set)) {
      i = add_client(unix_fd, &max_client, &client_set);
      add_connection(i);
    } /*if*/
  } /*while*/

}

static int
scan_int_or_char (unsigned char** p)
{
  if (**p == '\\') {
    *p = *p + 1;
    if (**p == 0) return -1;
    return (isdigit(**p) ? strtol(*p, (char**)p, 0) : *(*p)++);
  } else {
    return *(*p)++;
  } /*if*/
}

static int
process_inword (char* arg)
{
  char *p;
  int i;
  int from, to;
  static char default_word_chars[] = "_[0-9][A-Z][a-z][\\192-\\214][\\216-\\246][\\248-\\255]";

  p = (arg != 0 ? arg : default_word_chars);
  while (*p != 0) {
    if (*p == '[') {
      *p++;
      if (0 > (from = scan_int_or_char((unsigned char**)&p))) return -1;
      if (*p++ != '-') {
        return -1;
      }
      if (0 > (to = scan_int_or_char((unsigned char**)&p))) return -1;
      for (i = from; i <= to; i++) {
        set_bit(word_chars, i);
      } /*for*/
      if (*p++ != ']') return -1;
    } else {
      if (0 > (i = scan_int_or_char((unsigned char**)&p))) return -1;
      set_bit(word_chars, i);
    } /*if*/
  } /*while*/

  return 0;
}

static void
process_ignored_list (char* arg)
{
  struct stat st;
  char* p;
  int vt;
  int status;
  /* They can specify the terminal either with an absolute pathname, or a number. */
  /* Anything else is an error. */
  /* The number is what's passed to the server, so if they specify a pathname,
     we stat it and look at its device number. */
  
  p = strtok(arg, " \t\r\n");
  while (p != 0) {
    if (isdigit(p[0])) {
      vt = atoi(p);
    } else if (p[0] == '/') {
      status = stat(p, &st);
      if (status < 0) {
        perror(p);
        exit(3);
      } /*if*/
      if (major(st.st_rdev) != TTY_MAJOR) {
        fprintf(stderr, "%s: not a virtual terminal\n", p);
        exit(3);
      } /*if*/
      vt = minor(st.st_rdev);
    } else {
      print_docstrings(3);
    } /*if*/
    set_bit(ignored, vt);
    p = strtok(0, " \t\r\n");
  } /*while*/
}

int 
main (int argc, char** argv)
{
  static int opt_daemon = 1;
  static const char* mouse_type = "auto";
  static const char* device = "mouse";
  int c;
  char giiarg[100];
  char auto_sock_path[50];
  char repeater_path[50];
  char* my_sock_path;
  int unix_fd;
  struct sigaction sa;
  gii_input_t giin;
  int i;
  repeat_event_t event_fun;
  repeat_init_t init_fun;
  void* repeater_handle;

  giin = 0;
                                /* parse command line */
  parse_options(argc, argv);
  if (opt_foreground) opt_daemon = 0;
  if (opt_protocol != 0) mouse_type = opt_protocol;
  if (opt_device != 0) device = opt_device;
  if (opt_file != 0) {
    my_sock_path = opt_file;
  } else {
    if (0 > (c = snprintf(auto_sock_path, sizeof(auto_sock_path), gim_sock_path, device)) ||
        c >= sizeof(auto_sock_path))
      print_docstrings(2);
    my_sock_path = auto_sock_path;
  } /*if*/

  if (opt_repeater != 0) {
    repeater_handle = dlopen(opt_repeater, RTLD_NOW);
    if (repeater_handle == 0) {
      fprintf(stderr, "%s\n", dlerror());
      print_docstrings(2);
    } /*if*/
    event_fun = dlsym(repeater_handle, "repeat_event");
    if (event_fun == 0) {
      fprintf(stderr, "%s\n", dlerror());
      print_docstrings(2);
    } /*if*/
    init_fun = dlsym(repeater_handle, "repeat_init");
    if (init_fun == 0) {
      fprintf(stderr, "%s\n", dlerror());
      print_docstrings(2);
    } /*if*/
    
  } else {
    repeater_handle = 0;
    event_fun = 0;
    init_fun = 0;
  } /*if*/

  opt_debug = opt_verbose;
  if (opt_interval != 0) {
    double_time.tv_sec = 0;
    double_time.tv_usec = atoi(opt_interval);
  }
  if (opt_scale_down != 0) scale_down = atoi(opt_scale_down);
  if (opt_scale_up != 0) scale_up = atoi(opt_scale_up);
  if (opt_threshold != 0) threshold = atoi(opt_threshold);
  if (opt_age_limit != 0) max_selection_age = atoi(opt_age_limit);
  if (opt_buttons != 0) {
    for (i = 0; i < 9 && opt_buttons[i] != 0; i++) {
      if (!isdigit(opt_buttons[i])) print_docstrings(2);
      button_sequence[i] = opt_buttons[i] - '0';
    } /*for*/
  } /*if*/
  if (opt_ignore != 0) process_ignored_list(opt_ignore);
  if (process_inword(opt_wchars) < 0) print_docstrings(2);
  if (opt_version) print_version_and_copy();
  if (opt_help) print_docstrings(0);

  if (optind < argc) print_docstrings(2);

  openlog("gim", (opt_daemon ? LOG_PID : LOG_PERROR), LOG_DAEMON);
                                /* setup signal handling */
  sigemptyset(&sa.sa_mask);
  sa.sa_handler = handle_sig;
  sa.sa_flags = SA_INTERRUPT;
  sigaction(SIGINT, &sa, 0);
  sigaction(SIGTERM, &sa, 0);
  sigaction(SIGHUP, &sa, 0);
  sigaction(SIGPIPE, &sa, 0);

  giiInit();
  exc_init();
  unix_fd = -1;
                                /* fork and drop terminal */  
  exc_try {
    if (opt_daemon) {
      if (daemon(0,0) == -1) {
        int my_errno = errno;
        syslog(LOG_ERR, "daemon(0,0): %m");
        exc_eraise2(unix_error, my_errno);
      } /*if*/
      umask(0);
    } /*if*/

    /* set up word character table */
    load_inword(word_chars);

    unix_fd = open_unix(my_sock_path);

    /* open gii stream */ 
    if ((c = snprintf(giiarg, sizeof(giiarg), "linux-mouse:%s,%s/%s",
                      mouse_type, _PATH_DEV, device)) == -1 || c >= sizeof(giiarg)) {
      syslog(LOG_ERR, "buffer too small, bad mouse type?");
      exc_eraise(buffer_too_small);
    } /*if*/

    while (1) {
      if ((giin = giiOpen(giiarg)) == 0) {
        syslog(LOG_ERR, "cannot open gii stream, bad mouse type?");
        exc_eraise(gii_error);
      } /*if*/

      cook_state = 0; acx = acy = 0;

      if (init_fun != 0) {
        if (0 > (*init_fun)()) {
          int my_errno = errno;
          syslog(LOG_ERR, "repeater initialization error: %m");
          exc_eraise2(unix_error, my_errno);
        } /*if*/
      }

      exc_try {
        if (-1 == giiQueryDeviceInfoByNumber(giin, 0, 0, &gii_di)) {
          syslog(LOG_ERR, "cannot get mouse info");
          exc_eraise(gii_error);
        } /*if*/

        event_loop(giin, unix_fd, event_fun);
      }
      exc_finally {
        giiClose(giin);
      } exc_end;
    } /*while*/
  }
  exc_except(async_signal) {
    syslog(LOG_DEBUG, strsignal(*((int*)exc_info)));
  }
  exc_except(unix_error) {
    syslog(LOG_DEBUG, "%s: %d: %s", exc_file, exc_line, strerror(*((int*)exc_info)));
  }
  exc_except(exc_any) {
    syslog(LOG_DEBUG, "%s: %d: %s", exc_file, exc_line, exc_name);
  }  exc_end;

  if (unix_fd != -1) close(unix_fd);
  unlink(my_sock_path);
  giiExit();
  closelog();
  if (repeater_handle != 0) dlclose(repeater_handle);
  return 0;

}
